# Copyright 2024 CJVT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from stark.data.representation.tree import RepresentationTree


class QueryRepresentationTree(RepresentationTree):
    def __init__(self, node, children, filters):
        super().__init__(node, children)

    @classmethod
    def copy(cls, node, children, filters):
        """
        Copies original object with empty children.
        :param filters:
        :param children:
        :param node:
        :return:
        """
        return cls(node, children, filters)
